/******************************************************************************
 * lightmanager.cpp 
 * implements the management of all lights within the scene
 * Copyright (C) 2010 Marc Gissler <gisslerm@informatik.uni-freiburg.de>
 ******************************************************************************/

#include "lightmanager.h"

#include "shadow.h"

void CLight::setAmbientExclusively()
{
	GLfloat nullColor[3] = { 0.0f, 0.0f, 0.0f };
	glLightfv(m_lightId, GL_AMBIENT, m_ambientColor);
	glLightfv(m_lightId, GL_DIFFUSE, nullColor);
	glLightfv(m_lightId, GL_SPECULAR, nullColor);
}

void CLight::setAllIntensities()
{
	glLightfv(m_lightId, GL_AMBIENT, m_ambientColor);
	glLightfv(m_lightId, GL_DIFFUSE, m_diffuseColor);
	glLightfv(m_lightId, GL_SPECULAR, m_specularColor);
}

CPointLight::CPointLight(GLenum lightId, gmVector4& pos, gmVector3& ambient, gmVector3& diffuse, gmVector3& specular) 
{
	m_lightId = lightId;
	m_pos[0] = (GLfloat) pos.x; m_pos[1] = (GLfloat) pos.y; m_pos[2] = (GLfloat) pos.z; m_pos[3] = (GLfloat) pos.w;
	m_ambientColor[0] = (GLfloat) ambient.x; m_ambientColor[1] = (GLfloat) ambient.y; m_ambientColor[2] = (GLfloat) ambient.z;
	m_diffuseColor[0] = (GLfloat) diffuse.x; m_diffuseColor[1] = (GLfloat) diffuse.y; m_diffuseColor[2] = (GLfloat) diffuse.z;
	m_specularColor[0] = (GLfloat) specular.x; m_specularColor[1] = (GLfloat) specular.y; m_specularColor[2] = (GLfloat) specular.z;	
	m_castShadows = false;

	// communicate all settings to OpenGL
	glLightfv(m_lightId, GL_AMBIENT, m_ambientColor);
	glLightfv(m_lightId, GL_DIFFUSE, m_diffuseColor);
	glLightfv(m_lightId, GL_SPECULAR, m_specularColor);
	glLightfv(m_lightId, GL_POSITION, m_pos);
	glEnable(m_lightId);
}

void CPointLight::draw()
{
	if(m_pos[3] == 0.0) return;

	GLUquadric* quad;
	quad = gluNewQuadric();
	glPushMatrix();
	glTranslatef(m_pos[0], m_pos[1], m_pos[2]);
	gluSphere(quad, 0.1, 10, 10);
	glPopMatrix();
}

void CPointLight::position()
{
	glLightfv(m_lightId, GL_POSITION, m_pos);
}

CSpotLight::CSpotLight(GLenum lightId, gmVector4& pos, gmVector3& direction, float cutOff, gmVector3& ambient, gmVector3& diffuse, gmVector3& specular)
{
	m_lightId = lightId;
	m_pos[0] = (GLfloat) pos.x; m_pos[1] = (GLfloat) pos.y; m_pos[2] = (GLfloat) pos.z; m_pos[3] = (GLfloat) pos.w;
	m_dir[0] = (GLfloat) direction.x; m_dir[1] = (GLfloat) direction.y; m_dir[2] = (GLfloat) direction.z;
	m_cutOff[0] = cutOff;
	m_ambientColor[0] = (GLfloat) ambient.x; m_ambientColor[1] = (GLfloat) ambient.y; m_ambientColor[2] = (GLfloat) ambient.z;
	m_diffuseColor[0] = (GLfloat) diffuse.x; m_diffuseColor[1] = (GLfloat) diffuse.y; m_diffuseColor[2] = (GLfloat) diffuse.z;
	m_specularColor[0] = (GLfloat) specular.x; m_specularColor[1] = (GLfloat) specular.y; m_specularColor[2] = (GLfloat) specular.z;
	m_castShadows = false;

	// communicate all settings to OpenGL
	glLightfv(m_lightId, GL_AMBIENT, m_ambientColor);
	glLightfv(m_lightId, GL_DIFFUSE, m_diffuseColor);
	glLightfv(m_lightId, GL_SPECULAR, m_specularColor);
	glLightfv(m_lightId, GL_POSITION, m_pos);
	glLightfv(m_lightId, GL_SPOT_DIRECTION, m_dir);
	glLightfv(m_lightId, GL_SPOT_CUTOFF, m_cutOff);  
	glEnable(m_lightId);
}

void CSpotLight::draw()
{
	GLUquadric* quad;
	quad = gluNewQuadric();
	glPushMatrix();
	glTranslatef(m_pos[0], m_pos[1], m_pos[2]);
	gluSphere(quad, 0.1, 10, 10);
	glBegin(GL_LINES);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(m_dir[0], m_dir[1], m_dir[2]);
	glEnd();
	glPopMatrix();
}

void CSpotLight::position()
{
	glLightfv(m_lightId, GL_POSITION, m_pos);
	glLightfv(m_lightId, GL_SPOT_DIRECTION, m_dir);
}

// ----------------------------------------------------------------------------------------------------

CLightManager* CLightManager::m_pInstance = NULL; 

CLightManager::CLightManager() : m_pShadow(NULL)
{
	int tmp[1];
	glGetIntegerv(GL_MAX_LIGHTS, tmp);
	m_maxLights = tmp[0];

	clear();
}

CLightManager::~CLightManager()
{
	clear();

	if(m_pShadow)
		delete m_pShadow;
}

int CLightManager::addPointLight(gmVector4& pos, gmVector3& ambient, gmVector3& diffuse, gmVector3& specular)
{
	int m = m_maxLights / (int) 2;
	if((int)m_pointLights.size() >= m) return - 1;

	int lightId = GL_LIGHT0 + (int) m_pointLights.size();
	CPointLight* pl = new CPointLight(lightId, pos, ambient, diffuse, specular);
	m_pointLights.push_back(pl);
	return m_pointLights.size() - 1;
}

int CLightManager::addSpotLight(gmVector4& pos, gmVector3& direction, float cutOff, gmVector3& ambient, gmVector3& diffuse, gmVector3& specular)
{
	int m = m_maxLights / (int) 2;
	if((int)m_pointLights.size() >= m) return - 1;
	
	int lightId = GL_LIGHT0 + m + (int) m_pointLights.size();
	CSpotLight* sl = new CSpotLight(lightId, pos, direction, cutOff, ambient, diffuse, specular);
	m_spotLights.push_back(sl);
	return m + m_spotLights.size() - 1;
}

void CLightManager::clear()
{
	for(size_t i = 0; i < m_pointLights.size(); ++i)
		delete m_pointLights[i];
	m_pointLights.clear();
	
	for(size_t i = 0; i < m_spotLights.size(); ++i)
		delete m_spotLights[i];
	m_spotLights.clear();
}

void CLightManager::destroyLightManager()
{
	if(m_pInstance)
		delete m_pInstance;
	m_pInstance = NULL;
}

void CLightManager::drawSceneAlignedLights()
{
	for(size_t i = 0; i < m_pointLights.size(); ++i)
	{
		CPointLight* pl = m_pointLights[i];
		pl->draw();
	}

	for(size_t i = 0; i < m_spotLights.size(); ++i)
	{
		CSpotLight* sl = m_spotLights[i];
		sl->draw();
	}
}

void CLightManager::drawSceneWithShadows()
{
	m_pShadow->drawSceneWithShadows();
}

CLight*  CLightManager::getLight(int lightId)
{
	CLight* l = 0;
	int m = m_maxLights / 2;

	if(lightId >= m)
		l = m_spotLights[lightId - m];
	else
		l = m_pointLights[lightId];

	return l;
}

std::vector<CLight*> CLightManager::getLights()
{
	std::vector<CLight*> vecLights;
	for(size_t i = 0; i < m_pointLights.size(); ++i)
		vecLights.push_back(m_pointLights[i]);
	for(size_t i = 0; i < m_spotLights.size(); ++i)
		vecLights.push_back(m_spotLights[i]);

	return vecLights;
}

CLightManager* CLightManager::getLightManager()
{
	if(m_pInstance == NULL)
		m_pInstance = new CLightManager();
	return m_pInstance;
}

bool CLightManager::isLightEnabled(int lightId)
{
	return glIsEnabled(GL_LIGHT0+lightId);
}

void CLightManager::positionSceneAlignedLights()
{
	for(size_t i = 0; i < m_pointLights.size(); ++i)
	{
		CPointLight* pl = m_pointLights[i];
		pl->position();
	}

	for(size_t i = 0; i < m_spotLights.size(); ++i)
	{
		CSpotLight* sl = m_spotLights[i];
		sl->position();
	}
}

void CLightManager::setHasShadow(int lightId, bool hasShadow)
{
	int m = m_maxLights / 2;
	CLight* l = 0;
	if(lightId >= m)
		l = m_spotLights[lightId - m];
	else
		l = m_pointLights[lightId];

	l->m_castShadows = hasShadow;

	if(l->m_castShadows)
	{
		// instantiate one of the shadow algorithms
		if(m_pShadow == NULL)
		{
			//m_pShadow = CShadow::instantiateShadow("volume-shadow");
			m_pShadow = CShadow::instantiateShadow("shadow-map");
		}

		m_pShadow->build(lightId, gmVector3(l->m_pos[0], l->m_pos[1], l->m_pos[2]));
	}
}

void CLightManager::setAmbientExclusively()
{
	for(size_t i = 0; i < m_pointLights.size(); ++i)
	{
		CPointLight* pl = m_pointLights[i];
		pl->setAmbientExclusively();
	}

	for(size_t i = 0; i < m_spotLights.size(); ++i)
	{
		CSpotLight* sl = m_spotLights[i];
		sl->setAmbientExclusively();
	}
}

void CLightManager::setAllIntensities()
{
	for(size_t i = 0; i < m_pointLights.size(); ++i)
	{
		CPointLight* pl = m_pointLights[i];
		pl->setAllIntensities();
	}

	for(size_t i = 0; i < m_spotLights.size(); ++i)
	{
		CSpotLight* sl = m_spotLights[i];
		sl->setAllIntensities();
	}
}

void CLightManager::toggleLight(int lightId, bool toggle)
{
	//TODO: do not toggle directly, but notify the respective light!
	/*if(toggle)
		glEnable(GL_LIGHT0+lightId);
	else
		glDisable(GL_LIGHT0+lightId);*/
}
